module type ARRAY = sig
  (* two representation of an array:
   * 1. a list of elements
   * 2. an integer (a bit vector)
   *)
  type 'a t =
  | List of 'a list
  | Int of int

  val fromList : 'a list -> 'a t
  val fromInt : int -> bool t
  val getAny : int -> 'a t -> 'a
  val getBool : int -> bool t -> bool
end

module Array : ARRAY = struct
  type 'a t =
  | List of 'a list
  | Int of int

  let fromList l = List l
  let fromInt i = Int i

  (* This won't type check, but a readonable guess *)
  (* Why? In the Int branch we always returns a `bool`, but we are expecting an `'a` *)
  (* let getAny i = function
  | List l -> List.nth l i
  | Int n -> ((Int.shift_right n i) land 1) > 0 *)
  let getAny i = failwith "Cannot really implement this"

  let getBool i = function
  | List l -> List.nth l i
  | Int n -> ((Int.shift_right n i) land 1) > 0

end

(* Introducing GADT syntax *)
(* Normally we write *)
type 'a llist = Nil | Cons of 'a * 'a llist
type 'a option = None | Some of 'a

(* But we can also write *)
type 'a list' =
  | Nil' : 'a list'
  | Cons' : 'a * 'a list' -> 'a list'
type 'a option' =
  | None' : 'a option'
  | Some' : 'a -> 'a option'

(* But now it gives us some flexibility *)
type 'a array =
  | List : 'a list -> 'a array
  | Int : int -> bool array

let toString : char array -> string = function
| List l -> String.concat "" (List.map (String.make 1) l)
| _ -> .

(* List Frustrations *)
let head : 'a list -> 'a = function
| [] -> failwith "Empty list"
| x :: _ -> x

(* encode length at a type level *)
(* type ('a, 'len) list =
  | Nil : ('a, 0) list
  | Cons : 'a * ('a, 'len) list -> ('a, 'len + 1) list *)
(* Hummm, but 0 and 1 are not types *)
(* Or are they? *)
type z    = |
type 'n s = |

type ('a, 'len) list =
  | Nil : ('a, z) list
  | Cons : 'a * ('a, 'len) list -> ('a, 'len s) list

let head : ('a, 'len s) list -> 'a = function
| Cons (x, _) -> x
| _ -> .

(* A proof that our map function preserves length *)
let rec map : type n. ('a -> 'b) -> ('a, n) list -> ('b, n) list = fun f xs ->
  match xs with
  | Nil          -> Nil
  | Cons (x, xs) -> Cons (f x, map f xs)

(* Can we do? *)
(* let rec append : type n m. ('a, n) list -> ('a, m) list -> ('a, n + m) list = fun xs ys ->
  match xs with
  | Nil          -> ys
  | Cons (x, xs) -> Cons (x, append xs ys) *)
(* let rec filter : type n. ('a -> bool) -> ('a, n) list -> ('a, ?) list = fun f xs ->
  match xs with
  | Nil          -> Nil
  | Cons (x, xs) -> if f x then Cons (x, filter f xs) else filter f xs *)
(* with trick you can, but in general no *)
(* Forshadowing... *)

(* Red-black tree invariant *)
(*
 * 1. Every node is either red or black
 * 2. The emoty tree is black
 * 3. If a node is red, then both its children are black
 * 4. Every path from a node to any of its descendant leaf nodes contains the same number of black nodes
 * 5. All leaves are black
 *)

type red = |
type black = |
type ('a, 'color, 'n) tree =
| Empty : ('a, black, z) tree
| Red : 'a * ('a, black, 'n) tree * ('a, black, 'n) tree -> ('a, red, 'n s) tree
| Black : 'a * ('a, 'color1, 'n) tree * ('a, 'color2, 'n) tree -> ('a, black, 'n s) tree